function [res_direct, res_prob] = trainCDMCBoost(data_name, data, true_labels, n_classes,...
                   boosting_order, weak_learner_type, loss_type,...
                   code_type, user_codewords, exp_id, max_number_iterations,...
                   n_bins, n_extra_features, cost_matrix, step_size_search_type,...
                   data_test, true_labels_test, decision_rule_type)
               
%%%%%%%%%% Initialization%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% lower_cost = ones(n_classes) - eye(n_classes); %%%%% This is not final

file_name = ['classifiers/cdmcboost_' data_name '_id_'  num2str(exp_id) ];

if strcmp(code_type, 'simplex')
    codeword_set = buildSimplexCodewords(n_classes);
elseif strcmp(code_type, 'full')
    codeword_set = eye(n_classes);
elseif strcmp(code_type, 'full_zero')    
    codeword_set = eye(n_classes) - (ones(n_classes)/ n_classes);
elseif strcmp(code_type, 'costum')
    codeword_set = user_codewords;
end


if strcmp(loss_type , 'logExp')
    lower_cost = cost_matrix + eye(n_classes);
else
    lower_cost = cost_matrix;
end
    
if strcmp(loss_type , 'exp') & (floor(exp_id/100) == 2)
    lower_cost = ones(n_classes) - eye(n_classes);
end

if (floor(exp_id/100) == 7)
    lower_cost = ones(n_classes) - eye(n_classes);
end
    



start_iteration = 1;
n_samples = size(data, 1);
n_code_dimension = size(codeword_set, 1);

if strcmp(weak_learner_type,'regression')
    temp_node.feature_num = 1;
    temp_node.coeff_a = 0;
    temp_node.coeff_b = 0;
    temp_node.step_size = 0;
elseif strcmp(weak_learner_type,'stumps')
    temp_node.feature_num = 1;
    temp_node.thr = 0;
    temp_node.polarity = 0;
    temp_node.step_size = 0;
end
for i = 1 : n_code_dimension
    Classifier.dims(i).nodes(1) = temp_node;
end

Classifier.weak_learner_type = weak_learner_type;
Classifier.cost_matrix = cost_matrix;
Classifier.codeword_set = codeword_set;
Classifier.extra_features = [];
Classifier.has_extras = false;
if n_extra_features > 0
    n_data_dim = size(data,2 );
    extra_features = [eye(n_data_dim) rand(n_data_dim, n_extra_features)];
    Classifier.extra_features = extra_features;
    Classifier.has_extras = true;
    data = data * extra_features;
end

%%%%%%%%%% Preprocessing%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
lower_cost_matrix = lower_cost(true_labels, :);

repeated_codeworeds = zeros(n_samples, n_classes, n_code_dimension);
for i = 1 : n_code_dimension
    repeated_codeworeds(:,:, i) = repmat(codeword_set(i, :), n_samples,1);
end
repeated_codewords_square = repeated_codeworeds.^2;
     
if strcmp(weak_learner_type,'regression')
    data2 = data.^2;
elseif strcmp(weak_learner_type,'stumps')
    thr_set = buildThrSet(data, n_bins);
    index_map = buildIndexMap(data, thr_set);
end

%%%%%%%%%% CDMCBoost %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
predictor = cmpPredictor(Classifier, data);
projections = predictor * codeword_set;
[temp, predictions] = max(projections,[],2);

accuracy(start_iteration) = mean(predictions == true_labels);
risk(start_iteration) = cmpRisk(projections, true_labels,lower_cost_matrix, loss_type);
% disp([accuracy(start_iteration) risk(start_iteration)])

for i = start_iteration : max_number_iterations         
%     disp(['iteration: ' num2str(i)])
 
    for j = 1 : n_code_dimension
%         fprintf('%d ',j)
        grad_1 = cmpFirstGrad(j, projections, loss_type,...
                               repeated_codeworeds(:,:, j), lower_cost_matrix,...
                               codeword_set, true_labels);        
        if strcmp(boosting_order,'second')
            %%% will fix Later
            grad_2 = cmpSecondGrad(j, projections, loss_type, repeated_codeworeds(:,:, j), repeated_codewords_square(:,:, j));
        elseif strcmp(boosting_order,'first')
            grad_2 = ones(size(grad_1));
        end                
        grad_1 = grad_1 ./ grad_2;
        if strcmp(weak_learner_type,'regression')
            [update_node, update_angle, update_predictor] = findBestFeatureRegression(grad_1, grad_2, data, data2);
        elseif strcmp(weak_learner_type,'stumps')
            [update_node, update_angle, update_predictor] = findBestFeatureStump(data, index_map, grad_1, thr_set);
        end
        
%         update_predictor = grad_1;
        
        update_projection = update_predictor * codeword_set(j, :);
        
        if strcmp(step_size_search_type, 'linear')
            update_node.step_size = findBestStepSize(j, projections, update_projection,...
                lower_cost_matrix, loss_type,...
                repeated_codeworeds(:,:, j), true_labels);
        elseif strcmp(step_size_search_type, 'binary')
            update_node.step_size = findBestStepSizeBinary(j, projections, true_labels, update_projection, ...
                lower_cost_matrix, loss_type,...
                repeated_codeworeds(:,:, j), codeword_set);
        end

%         disp([update_node.step_size, bbt]);
%         input(':')

        Classifier.dims(j).nodes(i) = update_node; 
        predictor(:,j) = predictor(:,j) + update_node.step_size * update_predictor;       
        projections = projections + update_node.step_size * update_projection;
        [temp, predictions] = max(projections,[],2);
    end
%     fprintf('\n')
   
    risk(i) = cmpRisk(projections, true_labels,lower_cost_matrix, loss_type);
    accuracy(i) = mean (predictions == true_labels);
%     disp([accuracy(i) risk(i)]);
%     input(':')
 
    [str_loss_test(i) str_loss_bayes_test(i)] = cmpStrLoss(Classifier, data_test, true_labels_test, n_classes, decision_rule_type);
%     disp([str_loss_test(i) str_loss_bayes_test(i)]);   
 
%     save(file_name, 'Classifier', 'str_loss_test', 'str_loss_bayes_test');
%     disp('+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++')
end
save(file_name, 'Classifier', 'str_loss_test', 'str_loss_bayes_test');
res_direct = str_loss_test(end);
res_prob = str_loss_bayes_test(end);
